#!/usr/bin/python
# -*- coding: utf-8 -*-
 
 
"""
- on va rentrer un graphe comme un dictionnaire de dictionnaires pour pouvoir pondérer les arêtes : 
Pour chaque sommet, on donne les sommets adjacents et le poids associé. 
Par exemple : 

graphe = {'a': {'b': 4, 'c': 2},
     'b': {'a' : 4,'d': 5, 'c': 1},
     'c': {'a' : 2, 'b' : 1, 'd' : 8,'e': 10},
     'd': {'b': 5, 'c': 8, 'e': 2},
     'e': {'c': 10, 'd' : 2, 'f': 3},
     'f': {'e' : 3,'d' : 6}}

Commentaires pour les profs
Ainsi, g['a'] est le dictionnaire des voisins avec leurs poids :
 
>>> g['a']
{'c': 2, 'b': 4}
>>> g['a']['b']
4
>>> 'c' in g['a']
True
 
- pour trouver la clé correspondant à la valeur mini d'un dictionnaire, on utilise une syntaxe très pythonesque. Par exemple, pour avoir le voisin le plus proche de 'a' :
 
>>> min(g['a'], key = g['a'].get)
'c'
 
- dict.get(cle,defaut) : retourne la valeur de cle si elle se trouve dans le dictionnaire et defaut sinon ;
 
- float('inf') représente l'infini
 
"""
 
 
 
 
def affiche_peres(pere,depart,extremite,trajet):
    """
    À partir du dictionnaire des pères de chaque sommet on renvoie
    la liste des sommets du plus court chemin trouvé. Calcul récursif.
    On part de la fin et on remonte vers le départ du chemin.
    
    """
    if extremite == depart:
        return [depart] + trajet
    else:
        return (affiche_peres(pere, depart, pere[extremite], [extremite] + trajet))
 
def plus_court(graphe,etape,fin,visites,dist,pere,depart):
    """
    Trouve récursivement la plus courte chaine entre debut et fin avec l'algo de Dijkstra
    visites est une liste et dist et pere des dictionnaires 
    graphe  : le graphe étudié                                                               (dictionnaire)
    étape   : le sommet en cours d'étude                                                     (sommet)
    fin     : but du trajet                                                                  (sommet)
    visites : liste des sommets déjà visités                                                 (liste de sommets)
    dist    : dictionnaire meilleure distance actuelle entre départ et les sommets du graphe (dict sommet : int)
    pere    : dictionnaire des pères actuels des sommets correspondant aux meilleurs chemins (dict sommet : sommet)
    depart  : sommet global de départ                                                        (sommet)
    Retourne le couple (longueur mini (int), trajet correspondant (liste sommets)) 
       
    """
    # si on arrive à la fin, on affiche la distance et les peres
    if etape == fin:
       return dist[fin], affiche_peres(pere,depart,fin,[])
    # si c'est la première visite, c'est que l'étape actuelle est le départ : on met dist[etape] à 0
    if  len(visites) == 0 : dist[etape]=0
    # on commence à tester les voisins non visités
    for voisin in graphe[etape]:
        if voisin not in visites:
            # la distance est soit la distance calculée précédemment soit l'infini
            dist_voisin = dist.get(voisin,float('inf'))
            # on calcule la nouvelle distance calculée en passant par l'étape
            candidat_dist = dist[etape] + graphe[etape][voisin]
            # on effectue les changements si cela donne un chemin plus court
            if candidat_dist < dist_voisin:
                dist[voisin] = candidat_dist
                pere[voisin] = etape
    # on a regardé tous les voisins : le noeud entier est visité
    visites.append(etape)
    # on cherche le sommet *non visité* le plus proche du départ
    non_visites = dict((s, dist.get(s,float('inf'))) for s in graphe if s not in visites)
    noeud_plus_proche = min(non_visites, key = non_visites.get)
    # on applique récursivement en prenant comme nouvelle étape le sommet le plus proche 
    return plus_court(graphe,noeud_plus_proche,fin,visites,dist,pere,depart)
 
def dij_rec(graphe,debut,fin):
    return plus_court(graphe,debut,fin,[],{},{},debut)

### Corps du programme
'''on va rentrer un graphe comme un dictionnaire de dictionnaires pour pouvoir pondérer les arêtes : 
Pour chaque sommet, on donne les sommets adjacents et le poids associé. 
Par exemple : 
'''

#question II 2)
grapheQuestion2 = {
      'a': {'d' : 70, 'h' : 60, 'c' : 45, 'b': 30},
      'b': {'e' : 50, 'c' : 30, 'a' : 30, 'd': 80, 'f': 35},
      'c': {'b' : 30, 'a' : 45, 'g' : 90, 'h': 25},
      'd': {'e' : 60, 'f' : 35, 'a' : 70, 'b': 80},
      'e': {'d' : 60, 'b' : 50},
      'f': {'d' : 35, 'b' : 35},
      'g': {'h' : 40, 'i' : 20, 'c' : 90},
      'h': {'a' : 60, 'c' : 25, 'i' : 25, 'g': 40},
      'i': {'g' : 20, 'h' : 25},
}
longeur,chemin = dij_rec(grapheQuestion2,'g','d')
print('Plus court chemin : ',chemin, ' de longueur :',longeur)

#question II 3)
grapheQuestion3 = {
      'a': {'d' : 70, 'h' : 60, 'c' : 45, 'b': 30},
      'b': {'e' : 50, 'c' : 30, 'a' : 30, 'd': 80, 'f': 35},
      'c': {'b' : 30, 'a' : 45, 'g' : 90},
      'd': {'e' : 60, 'f' : 35, 'a' : 70, 'b': 80},
      'e': {'d' : 60, 'b' : 50},
      'f': {'d' : 35, 'b' : 35},
      'g': {'h' : 40, 'i' : 20, 'c' : 90},
      'h': {'a' : 60, 'i' : 25, 'g': 40},
      'i': {'g' : 20, 'h' : 25},
}
longeur,chemin = dij_rec(grapheQuestion3,'g','d')
print('Plus court chemin : ',chemin, ' de longueur :',longeur)